#!/usr/bin/env python
#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import sys


def main() -> int:
    """
    Entrypoint for ``nait`` which synthesizes a pytest for each fixture name as an function
    asserting success on the fixture's gather method. For example, given ::

        nait nanaimo_bar

    ... the following, equivalent test function would be synthesized ::

        test_nanaimo_bar(pytest_args, nanaimo_bar):
            assert_success(nanaimo_bar.gather(**pytest_args))

    """
    if '--environ-shell' in sys.argv or '-S' in sys.argv:
        import nanaimo
        defaults = nanaimo.config.ArgumentDefaults.create_defaults_with_early_rc_config()
        args_ns = nanaimo.Namespace(defaults=defaults, allow_none_values=False)
        for key, value in args_ns.get_as_merged_dict('environ').items():
            print('export {}="{}";'.format(key, value), end='')
        argv_len = len(sys.argv)
        for x in range(0, argv_len):
            if sys.argv[x] == '--environ' and x < argv_len - 1:
                environ_pair = sys.argv[x+1].split('=')
                if len(environ_pair) == 2:
                    print('export {}="{}";'.format(environ_pair[0].strip(), environ_pair[1].strip()))
        return 0
    if '--version' in sys.argv:
        from nanaimo.version import __version__
        print(__version__)
        return 0
    import nanaimo.pytest.plugin
    import pytest
    nanaimo.pytest.plugin._nait_mode = True
    return pytest.main()



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#
#                                       (@@@@%%%%%%%%%&@@&.
#                              /%&&%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%&@@(
#                              *@&%%%%%%%%%&&%%%%%%%%%%%%%%%%%%&&&%%%%%%%
#                               @   @@@(@@@@%%%%%%%%%%%%%%%%&@@&* @@@   .
#                               ,   .        .  .@@@&                   /
#                                .       .                              *
#                               @@              .                       @
#                              @&&&&&&@. .    .                     *@%&@
#                              &&&&&&&&&&&&&&&&@@        *@@############@
#                     *&/ @@ #&&&&&&&&&&&&&&&&&&&&@  ###################*
#                              @&&&&&&&&&&&&&&&&&&##################@
#                                 %@&&&&&&&&&&&&&&################@
#                                        @&&&&&&&&&&%#######&@%
#  nanaimo                                   (@&&&&####@@*
#
import tempfile
import textwrap
import typing

import nanaimo
import nanaimo.fixtures
import functools


class ProgramUploader(nanaimo.fixtures.SubprocessFixture):
    """
    JLinkExe fixture that loads a given hexfile to a target device.
    """

    fixture_name = 'nanaimo_jlink_upload'
    argument_prefix = 'jlink_up'

    script_template = textwrap.dedent('''
        f
        device {device}
        speed {speed}
        si {serial_interface}
        connect
        hwinfo
        st
        h
        moe
        rx {reset_delay_millis}
        loadfile {hexfile}
        r
        qc
        ''').lstrip()

    @classmethod
    def on_visit_test_arguments(cls, arguments: nanaimo.Arguments) -> None:
        super().on_visit_test_arguments(arguments)
        arguments.add_argument('--exe',
                               default='JLinkExe',
                               help='A path to the jlink commander executable.')
        arguments.add_argument('--hexfile',
                               help='Path to a hex file to upload.')
        arguments.add_argument('--device',
                               help='The target device.')
        arguments.add_argument('--interface-speed-khz',
                               help='The interface speed in kilohertz',
                               default='default')
        arguments.add_argument('--serial-interface',
                               help='The serial interface to use (e.g. swd, jtag)',
                               default='swd')
        arguments.add_argument('--reset-delay-millis',
                               default=400,
                               help='Milliseconds to wait after reset before starting to load the new image.')
        arguments.add_argument('--script',
                               default=None,
                               help=textwrap.dedent('''
                            Path to a jlink script to use. If not provided then an script will be generated using an
                            internal default. The following template parameters are supported but optional:

                                {device}             = The target device.
                                {speed}              = The interface speed in kilohertz
                                {serial_interface}   = The serial interface to use.
                                {reset_delay_millis} = An optional delay to include after reset.
                                {hexfile}            = The hexfile to upload.
                            ''').lstrip())

    def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
        """
        Construct a command to upload (aka "flash") a firmware to a target device using Segger's JLink
        Commander program with the assumption that a Segger debug probe like the JLink is attached to the
        system.
        +----------------+---------------------------------------+--------------------------------------------------+
        | **Artifacts**                                                                                             |
        |                                                                                                           |
        +----------------+---------------------------------------+--------------------------------------------------+
        | key            | type                                  | Notes                                            |
        +================+=======================================+==================================================+
        | ``tmpfile``    | Optional[tempfile.NamedTemporaryFile] | A temporary file containing the jlink script to  |
        |                |                                       | execute with expanded template parameters.       |
        +----------------+---------------------------------------+--------------------------------------------------+
        | ``scriptfile`` | Optional[str]                         | A path to a script file used to invoke jlink.    |
        |                |                                       | This file can contain template parameters.       |
        +----------------+---------------------------------------+--------------------------------------------------+

        """
        script_file_path = self.get_arg_covariant(arguments, 'script')
        tmpfile = tempfile.NamedTemporaryFile(mode='w')
        setattr(inout_artifacts, 'tmpfile', tmpfile)

        # without a script, other args are required:
        get_arg_func = functools.partial(self.get_arg_covariant_or_fail, arguments)

        if script_file_path is None:
            # keep around for as long as the command exists.
            setattr(inout_artifacts, 'scriptfile', None)
            template = self.script_template

        else:
            setattr(inout_artifacts, 'scriptfile', script_file_path)
            with open(script_file_path, 'r') as user_script_file:
                template = user_script_file.read()
            get_arg_func = functools.partial(self.get_arg_covariant, arguments)  # don't fail fixture on missing args

        # poor man's templating
        expanded_script_file = template.format(
                hexfile=get_arg_func('hexfile'),
                device=get_arg_func('device'),
                speed=get_arg_func('interface-speed-khz'),
                serial_interface=get_arg_func('serial-interface'),
                reset_delay_millis=str(get_arg_func('reset-delay-millis'))
            )

        with open(tmpfile.name, 'w') as tempfile_handle:
            tempfile_handle.write(expanded_script_file)

        return '{} -CommanderScript {}'.format(self.get_arg_covariant(arguments, 'exe', 'JLinkExe'),
                                               tmpfile.name)


def pytest_nanaimo_fixture_type() -> typing.Type['nanaimo.fixtures.Fixture']:
    return ProgramUploader



#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import pathlib
import re
import typing

import pytest

import nanaimo
from nanaimo import assert_success


@pytest.mark.asyncio
@pytest.mark.parametrize('use_internal_template', [True, False])
async def test_jlink_upload(use_internal_template: bool,
                            nanaimo_jlink_upload: nanaimo.fixtures.Fixture,
                            mock_JLinkExe: pathlib.Path,
                            test_build_config_hex: pathlib.Path,
                            test_jlink_template: pathlib.Path) -> None:
    """
        Test using the jlink ProgramUploader fixture.
    """
    test_args = {
        'jlink_up_exe': str(mock_JLinkExe),
        'jlink_up_hexfile': str(test_build_config_hex),
        'jlink_up_device': 'fake'}

    if not use_internal_template:
        test_args['jlink_up_script'] = str(test_jlink_template)
        test_args['jlink_up_device'] = None  # fails unless args are optional with script

    artifacts = assert_success(await nanaimo_jlink_upload.gather(**test_args))

    cmd = artifacts.cmd
    assert cmd is not None
    print(cmd)

    tmpfile = artifacts.tmpfile

    assert tmpfile is not None
    scriptfile = tmpfile.name
    assert pathlib.Path(scriptfile).exists()

    print(scriptfile)

    hexfile_in_script = None  # type: typing.Optional[str]
    with open(scriptfile, 'r') as scriptfile_handle:
        for line in scriptfile_handle:
            matchobj = re.search(r'^loadfile\s+((?:/|\w)+(?:/|[\w\. -])*\.hex)$', line)
            if matchobj:
                hexfile_in_script = matchobj.group(1)
            print(line, end='')

    assert hexfile_in_script is not None
    assert hexfile_in_script == str(test_build_config_hex)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# This software is distributed under the terms of the MIT License.
#

import pathlib
import re

import pytest

import nanaimo
import nanaimo.fixtures


@pytest.mark.asyncio
async def test_subprocess_match_filter(mock_JLinkExe: pathlib.Path) -> None:
    """
    Executing JLinkExe breaks subprocess output stream filtering.
    We replicate this by searching the output for a simple error
    which appears in stdout from JLinkExe when run from a shell, but
    is missing when run as a SubprocessFixture command:
    """
    class TestSubprocessFixture(nanaimo.fixtures.SubprocessFixture):
        argument_prefix = 'test'

        def on_construct_command(self, arguments: nanaimo.Namespace, inout_artifacts: nanaimo.Artifacts) -> str:
            return '{} --simulate-error'.format(str(mock_JLinkExe))

    subprocess_fixture = TestSubprocessFixture(nanaimo.fixtures.FixtureManager())
    filter = nanaimo.fixtures.SubprocessFixture.SubprocessMessageMatcher(re.compile('Simulating abnormal exit.'))
    subprocess_fixture.stdout_filter = filter
    subprocess_fixture.stderr_filter = filter
    await subprocess_fixture.gather(test_log_stdout=True)
    assert filter.match_count == 1
